#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
#
# This script is licensed under GNU GPL version 2.0 or above
# (c) 2025 Antonio J. Delgado
"""Scan Github packages"""

import sys
import os
import logging
from logging.handlers import SysLogHandler
import json
import re
import time
import click
import click_config_file
import requests


HOME_FOLDER = os.environ.get('HOME', os.environ.get('USERPROFILE', '/'))
if HOME_FOLDER == '/':
    CACHE_FOLDER = '/var/cache'
    LOG_FOLDER = '/var/log/'
else:
    CACHE_FOLDER = f"{HOME_FOLDER}/.local/"
    LOG_FOLDER = f"{HOME_FOLDER}/log/"


class GithubPackages:
    '''Obtain all Github Packages'''

    def __init__(self, **kwargs):
        self.config = kwargs
        if 'log_file' not in kwargs or kwargs['log_file'] is None:
            self.config['log_file'] = os.path.join(
                os.environ.get(
                    'HOME',
                    os.environ.get(
                        'USERPROFILE',
                        os.getcwd()
                    )
                ),
                'log',
                'github_packages.log'
            )
        self._init_log()
        self.session = requests.Session()
        self._default_data = {
            "last_update": 0,
            "orgs_with_assets": [],
            "orgs_without_assets": [],
            # "assets": [],
            "packages_urls": [],
            'organizations': [],
        }
        self._default_payload = {
                'per_page': 100,
            }
        self.request_count = {}
        headers = {
            'User-Agent': 'github_packages/0.0.1',
            'Accept': 'application/vnd.github+json',
            'Authorization': f"Bearer {self.config['github_token']}",
            'X-GitHub-Api-Version': '2022-11-28',
        }
        self.session.headers.update(headers)
        self.data = self._read_cached_data()
        self._get_github_rate_limit()

        if self.data['orgs_with_assets']:
            for organization in self.data['orgs_with_assets']:
                self._process_organization(organization)
        if (
            'organizations' in self.data and
            'organizations_last_update' in self.data and
            self.data['organizations_last_update'] + self.config['max_cache_age'] < time.time()
        ):
            organizations = self.data['organizations']
            self._log.debug(
                "A total of %s organizations existing in cache",
                len(organizations)
            )
        else:
            self._log.debug(
                "Getting all (or up to %s) organizations...",
                self.config['max_organizations'],
            )
            full_organizations = self._get_paginated_url(
                'https://api.github.com/organizations',
                self.config['max_organizations'],
            )
            organizations = []
            self._log.debug(
                "Obtained %s organizations",
                len(full_organizations)
            )
            for organization in full_organizations:
                if 'login' in organization:
                    organizations.append(organization['login'])
                elif 'message' in organization:
                    self._log.debug(
                        "Incomplete list of organizations. %s",
                        organization
                    )
                else:
                    self._log.error(
                        "Organization is not a dictionay? %s",
                        organization,
                    )
                    # sys.exit(2)
            self._log.debug(
                "A total of %s organizations fetched",
                len(organizations)
            )
            self.data['organizations_last_update'] = time.time()
            self._save_cached_data(self.data)
        for organization in organizations:
            self.data['organizations'].append(organization)
            try:
                self._process_organization(organization)
            except Exception as error:
                self._log.error(
                    "Error processing organization. Organization: %s. Error: %s",
                    organization,
                    error
                )

    def close(self, error=0):
        '''Close class and save data'''
        self._save_cached_data(self.data)
        if error > 0:
            sys.exit(error)

    def _process_organization(self, organization):
        self._log.debug(
            "Processing organization '%s'...",
            organization
        )
        if organization in self.data['orgs_without_assets']:
            self._log.debug(
                "Skipping organization '%s' that previously didn't have any assets",
                organization
            )
        else:
            self._log.debug(
                "Getting all (or up to %s) repositories from the organization '%s'...",
                self.config['max_repos_per_org'],
                organization
            )
            repositories = self._get_paginated_url(
                f"https://api.github.com/orgs/{organization}/repos",
                self.config['max_repos_per_org']
            )
            self._log.debug(
                "A total of %s repositories fetched",
                len(repositories)
            )
            for repository in repositories:
                try:
                    self._process_repository(repository)
                except Exception as error:
                    self._log.error(
                        "Error processing repository '%s'. %s",
                        repository,
                        error
                    )

    def _process_repository(self, repository):
        self._log.debug(
            "Getting latest release of the repo '%s'...",
            repository['name']
        )
        latest_release = self._get_paginated_url(
            f"{repository['url']}/releases/latest",
            max_items=1 # It should be onely one 'latest'
        )
        organization = repository['owner']
        if (
            'status' not in latest_release or
            latest_release['status'] != '404'
        ) and (
            'assets' in latest_release
            ):
            # self._log.debug(
            #     "Latest release: %s",
            #     latest_release
            # )
            # self.data['assets'] += latest_release['assets']
            # self._log.debug(
            #     "A total of %s assets",
            #     len(self.data['assets'])
            # )
            if len(latest_release['assets']) > 0:
                if organization['login'] not in self.data['orgs_with_assets']:
                    self.data['orgs_with_assets'].append(organization['login'])
                    if organization['login'] in self.data['orgs_without_assets']:
                        self.data['orgs_without_assets'].remove(organization['login'])
            else:
                if organization['login'] not in self.data['orgs_without_assets']:
                    if organization['login'] not in self.data['orgs_with_assets']:
                        self._log.debug(
                            "Organization '%s' doesn't have any assets, adding to skip list",
                            organization['login']
                        )
                        self.data['orgs_without_assets'].append(organization['login'])
            for asset in latest_release['assets']:
                self._process_asset(asset)
            self._save_cached_data(self.data)

    def _process_asset(self, asset):
        name_match = re.search(self.config['asset_name_regexp'], asset['name'])
        if name_match:
            self._log.debug(
                "Found a package that match '%s'. %s. %s",
                self.config['asset_name_regexp'],
                asset['name'],
                asset['browser_download_url']
            )
            self.data['packages_urls'].append(asset['browser_download_url'])


    def _read_cached_data(self):
        if os.path.exists(self.config['cache_file']):
            with open(self.config['cache_file'], 'r', encoding='utf-8') as cache_file:
                try:
                    cached_data = json.load(cache_file)
                    if (
                        'last_update' in cached_data and
                        cached_data['last_update'] + self.config['max_cache_age'] > time.time()
                    ):
                        cached_data = self._default_data
                except json.decoder.JSONDecodeError:
                    cached_data = self._default_data
                return cached_data
        else:
            return self._default_data

    def _save_cached_data(self, data):
        data['last_update'] = time.time()
        with open(self.config['cache_file'], 'w', encoding='utf-8') as cache_file:
            json.dump(data, cache_file, indent=2)

    def _new_request(self, headers):
        if 'X-RateLimit-Remaining' in headers:
            if int(headers['X-RateLimit-Remaining']) % 1000 == 0:
                self._log.debug(
                    "Rate limit remaining requests %s",
                    headers['X-RateLimit-Remaining']
                )
            if int(headers['X-RateLimit-Remaining']) < 50:
                if 'X-RateLimit-Reset' in headers:
                    wait = int(headers['X-RateLimit-Reset']) - time.time() + 5
                    str_reset_time = time.strftime(
                        "%Y/%m/%d %H:%M:%S %z",
                        time.gmtime(int(headers['X-RateLimit-Reset']))
                    )
                    if wait > 60:
                        pretty_wait = f"{round(wait / 60, 2)} minutes (until {str_reset_time})"
                    else:
                        pretty_wait = f"{wait} seconds (until {str_reset_time})"
                    self._log.warning(
                        "Rate limit too close to be consumed (%s requests remaining). Waiting %s until it's reset",
                        headers['X-RateLimit-Remaining'],
                        pretty_wait
                    )
                    time.sleep(wait)
        if self.request_count['time_start'] + 60*60 > time.time():
            self.request_count['max_per_minute'] -= 1
        if self.request_count['max_per_minute'] < 1:
            self._save_cached_data(self.data)
            self._log.warning(
                "Maximun number of request per minute (%s) reached, waiting 30 seconds",
                round(self.config['rate_limit'], 2)
            )
            time.sleep(30)
            self.request_count['max_per_minute'] = self.config['rate_limit']
            self.request_count['time_start'] = time.time()

    def _get_github_rate_limit(self):
        result = self.session.get("https://api.github.com/user")
        if 'X-RateLimit-Limit' in result.headers:
            self.config['rate_limit'] = (int(result.headers['X-RateLimit-Limit']) / 60) - 30
        else:
            self.config['rate_limit'] = 50
        self.request_count = {
            "time_start": time.time(),
            "max_per_minute": self.config['rate_limit']
        }
        self._new_request(result.headers)

    def _get_paginated_url(self, url, max_items, payload=None):
        if payload is None:
            payload = self._default_payload
        result = self.session.get(url, params=payload)
        self._new_request(result.headers)
        data = result.json()
        if isinstance(data, dict):
            items = [
                data
            ]
        else:
            items = data
        if 'status' in items:
            if int(items['status']) > 299 and items['status'] != '404':
                self._log.error(
                    "Error %s getting URL '%s'. Items: %s. Headers: %s",
                    items['status'],
                    url,
                    items,
                    result.headers
                )
                self.close(1)
            elif items['status'] == '404':
                self._log.debug(
                    "Not found any resource (404) in the endpoint '%s'.",
                    url
                )
            else:
                self._log.debug(
                    "Got a status field in the response. Response: %s",
                    items
                )
        page = 0
        next_link = self._get_next_pagination_link(result.headers)
        while next_link and len(items) < max_items:
            page += 1
            self._log.debug(
                "Obtained %s items so far",
                len(items)
            )
            result = self.session.get(next_link, params=payload)
            self._new_request(result.headers)
            result_data = result.json()
            if 'status' not in result_data:
                items += list(result_data)
            else:
                self._log.warning(
                    "Didn't obtain a successful response. Stopping fetching pages. %s",
                    result_data
                )
                return items
            next_link = self._get_next_pagination_link(result.headers)
        return items

    def _get_next_pagination_link(self, headers):
        if 'link' in headers:
            link_header = headers['Link']
            for link in link_header.split(', '):
                if link:
                    link_split = link.split('; ')
                    if link_split[1] == 'rel="next"':
                        return link_split[0].replace('<', '').replace('>', '')
        return None

    def _init_log(self):
        ''' Initialize log object '''
        self._log = logging.getLogger("github_packages")
        self._log.setLevel(logging.DEBUG)

        sysloghandler = SysLogHandler()
        sysloghandler.setLevel(logging.DEBUG)
        self._log.addHandler(sysloghandler)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(
            logging.getLevelName(self.config.get("debug_level", 'INFO'))
        )
        self._log.addHandler(streamhandler)

        if 'log_file' in self.config:
            log_file = self.config['log_file']
        else:
            log_file = os.path.join(LOG_FOLDER, "github_packages.log")

        if not os.path.exists(os.path.dirname(log_file)):
            os.mkdir(os.path.dirname(log_file))

        filehandler = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=102400000
        )
        # create formatter
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
        )
        filehandler.setFormatter(formatter)
        filehandler.setLevel(logging.DEBUG)
        self._log.addHandler(filehandler)
        return True

@click.command()
@click.option(
    "--debug-level",
    "-d",
    default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        case_sensitive=False,
    ),
    help='Set the debug level for the standard output.'
)
@click.option(
    '--max-organizations',
    '-m',
    default=200,
    help='Maximun number of organizations to fetch'
)
@click.option(
    '--max-repos-per-org',
    '-M',
    default=200,
    help='Maximun number of repositories per organization'
)
@click.option(
    '--log-file',
    '-l',
    default=f"{LOG_FOLDER}/github_packages_data.log",
    help="File to store all debug messages."
)
# @click.option("--dummy","-n", is_flag=True,
# help="Don't do anything, just show what would be done.")
@click.option(
    '--github-token',
    '-t',
    required=True,
    help='Github token with read access'
)
@click.option(
    '--cache-file',
    '-f',
    default=f"{CACHE_FOLDER}/github_packages_data.json",
    help='Cache file to store data from each run',
)
@click.option(
    '--asset-name-regexp',
    '-r',
    default=r'\.deb$',
    help='Regular expression to search in an assets name',
)
@click.option(
    '--max-cache-age',
    '-a',
    default=60*60*24*7,
    help='Max age in seconds for the cache'
)
@click_config_file.configuration_option()
def __main__(**kwargs):
    obj = GithubPackages(**kwargs)
    obj.close()

if __name__ == "__main__":
    __main__()
